using System;
using System.Collections;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using AspectLib;
using Rail.Reflect;

namespace AspectNet.AOP
{
	/// <summary>
	/// Summary description for Aspect.
	/// </summary>
	public class Aspect : ArrayList
	{
		private RAssemblyDef	_aspectAssembly = null;
		private RModuleDef		_aspectModule = null;
		private RTypeDef		_aspectType = null;

		private Hashtable		_name2AdviceMap = new Hashtable();
		
		/// <summary>
		/// Default constructor
		/// </summary>
		public Aspect( RAssemblyDef aspectAssembly,
						RModuleDef aspectModule,
						RTypeDef aspectType )
		{
			_aspectAssembly = aspectAssembly;
			_aspectModule = aspectModule;
			_aspectType = aspectType;

			// now load all the advices
			InitializeAdvices();
		}

		private void InitializeAdvices()
		{
			// loop over all methods
			foreach ( RMethodDef advice in _aspectType.GetMethods() )
			{
				// get the custom attributes of type AdviceAttribute
				RCustomAttrib[] adviceAttributes = 
					advice.GetCustomAttributes( typeof( AdviceAttribute ).FullName, false );

				// check if this method is an advice (has AdviceAttribute defined)
				if ( adviceAttributes != null && adviceAttributes.Length > 0 )
				{
					// it's guaranteed that we have:
					//   - 1 attribute
					//   - first property of it is the Name one
					string advName = adviceAttributes[0].GetPropertiesValues()[0].ToString();

					Advice adv = new Advice( this, advice, advName );

					this.Add( adv );
					_name2AdviceMap[ advName ] = adv;
				}
			}
		}

		public RAssemblyDef AspectAssembly
		{
			get { return _aspectAssembly; }
		}

		public RModuleDef AspectModule
		{
			get { return _aspectModule; }
		}

		public RTypeDef AspectType
		{
			get { return _aspectType; }
		}

		/// <summary>
		/// Returns a particular advice given its name
		/// </summary>
		public Advice this[string adviceName]
		{
			get
			{
				Advice result = null;

				if ( _name2AdviceMap.ContainsKey( adviceName ) )
				{
					result = (Advice)_name2AdviceMap[ adviceName ]	;
				}

				return result;
			}
		}
	}
}
