using System;
using System.Collections;

namespace AspectNet
{
	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	public class JPFactory
	{
		#region private section

		private Hashtable _jpBuilders				= new Hashtable();
		private Hashtable _jpWeavers				= new Hashtable();
		
		private static JPFactory _factoryInstance	= new JPFactory();

		private JPFactory()
		{
			RegisterJPBuilders();
			RegisterJPWeavers();
		}
		#endregion

		#region extension points
		
		/// <summary>
		/// Private method to create and register all the join point builders
		/// </summary>
		private void RegisterJPWeavers()
		{
			_jpWeavers.Add( typeof(MethodJPBuilder), new MethodJPWeaver() );
		}

		/// <summary>
		/// Private method to create and register all the join point builders
		/// </summary>
		private void RegisterJPBuilders()
		{
			IJPBuilder builder2Add = new MethodJPBuilder();
			
			_jpBuilders.Add( builder2Add.XmlTag, builder2Add );
		}

		#endregion

		#region public interface
		public static JPFactory Instance
		{
			get { return _factoryInstance; }
		}


		public IJPWeaver GetWeaver( JoinPoint jp )
		{
			IJPWeaver res = null;
			
			if ( _jpWeavers.ContainsKey( jp.JPBuilderType ) )
			{
				res = (IJPWeaver)_jpWeavers[jp.JPBuilderType];
			}
			
			return res;
		}

		public Hashtable JPBuilders
		{
			get { return _jpBuilders; }
		}

		public IJPBuilder GetBuilder( string xmlTag )
		{
			IJPBuilder res = null;
			
			if ( _jpBuilders.ContainsKey( xmlTag ) )
			{
				res = (IJPBuilder)_jpBuilders[xmlTag];
			}
			
			return res;
		}

		public void Reset()
		{
			foreach ( IJPWeaver weaver in this._jpWeavers.Values )
				weaver.Reset();

			foreach ( IJPBuilder builder in this._jpBuilders.Values )
				builder.Reset();
		}
		
		#endregion

	}
}
