using System;
using System.Reflection;
using AspectLib;
using Rail.Reflect;

namespace AspectNet
{
	/// <summary>
	/// Summary description for JoinPoint.
	/// </summary>
	public class JoinPoint
	{
		private RMethodDef		_methodInfo			= null;
		private RTypeDef		_declaringType		= null;
		private RModuleDef		_declaringModule	= null;
		private RAssemblyDef	_declaringAssembly	= null;

		private PointCut		_pointcut			= null;

		private Type			_pointType			= null;

		/// <summary>
		/// ctor
		/// </summary>
		public JoinPoint(PointCut pCut, RMethodDef methDef, Type pType)
		{
			_pointcut			= pCut;
			_methodInfo			= methDef;
			_pointType			= pType;
			
			_declaringType		= (RTypeDef)_methodInfo.DeclaringType;
			_declaringModule	= (RModuleDef)_declaringType.Module;
			_declaringAssembly	= (RAssemblyDef) _declaringModule.Assembly;

        }

		/// <summary>
		/// Returns the type of the IJPBuilder that matched the JP
		/// </summary>
		public Type JPBuilderType
		{
			get { return _pointType; }
		}


		/// <summary>
		/// Returns the parent pointcut
		/// </summary>
		public PointCut PointCut
		{
			get { return _pointcut; }
		}

		/// <summary>
		/// Returns the method definition
		/// </summary>
		public RMethodDef MethodDef
		{
			get { return _methodInfo; }
		}

		/// <summary>
		/// Returns the type that contains the joinpoint
		/// </summary>
		public RTypeDef DeclaringType
		{
			get { return _declaringType; }
		}

		/// <summary>
		/// Returns the module that contains the joinpoint
		/// </summary>
		public RModuleDef DeclaringModule
		{
			get { return _declaringModule; }
		}

		/// <summary>
		/// Returns the assembly that contains the joinpoint
		/// </summary>
		public RAssemblyDef DeclaringAssembly
		{
			get { return _declaringAssembly; }
		}
	}
}
