using System;
using System.Collections;
using AspectLib;
using Rail.Reflect;

namespace AspectNet
{
	/// <summary>
	/// Summary description for PointCut.
	/// </summary>
	public class PointCut : ArrayList
	{
		private string			_aspectName		= null;
		private string			_adviceName		= null;
		private BindingOpt		_bindOptions	= BindingOpt.bindNone;

		/// <summary>
		/// ctor
		/// </summary>
		public PointCut( string aspectName, 
							string adviceName, 
							BindingOpt bindOption )
		{
			_aspectName = aspectName;
			_adviceName = adviceName;
			_bindOptions = bindOption;
		}

		/// <summary>
		/// Assembly definition in which the joinpoints have been found
		/// </summary>
		public RAssemblyDef ParentAssembly
		{
			get
			{
				RAssemblyDef def = null;

				if ( this.Count > 0 )
				{
					def = (this[0] as JoinPoint).DeclaringAssembly;
				}

				return def;
			}
		}

		/// <summary>
		/// Get name of the aspect to which the joinpoints should be woven
		/// </summary>
		public string AspectName
		{
			get { return _aspectName; }
		}

		/// <summary>
		/// The name of the advice to which the joinpoints should be woven
		/// </summary>
		public string AdviceName
		{
			get { return _adviceName; }
		}

		/// <summary>
		/// The type of binding that should be used when weaving
		/// </summary>
		public BindingOpt Binding
		{
			get { return _bindOptions; }
		}
	}
}
