using System;
using System.Xml.Schema;

namespace AspectNet
{
	public class ErrorEventArgs : EventArgs
	{
		public enum ErrType
		{
			Error = 0,
			Warning = 1
		}

		public enum ErrKind
		{
			Unknown = 1	,
			ErrSchema,
			ErrMapping,
		}

		private ErrType _errorType;
		private ErrKind _errorKind;
		private object _origSender;

		private ValidationEventArgs _valArgs;
		
		private Exception _origException;
		private string _additionalMsg;


		public ErrorEventArgs( ErrKind errType, object sender, ValidationEventArgs vargs )
		{
			_errorKind = errType;
			_origSender = sender;
			_valArgs = vargs;
			_errorType = ErrType.Error;
		}

		public ErrorEventArgs( Exception origException )
		{
			_origException = origException;
			_errorKind = ErrKind.Unknown;
			_errorType = ErrType.Error;
		}

		public ErrorEventArgs( Exception origException, ErrType errType )
		{
			_origException = origException;
			_errorKind = ErrKind.Unknown;
			_errorType = errType;
		}

		public ErrorEventArgs( string warningTxt )
		{
			_additionalMsg = warningTxt;
			_errorKind = ErrKind.Unknown;
			_errorType = ErrType.Error;
		}

		public ErrType ErrorType
		{
			get { return _errorType; }
		}

		public ErrKind ErrorKind
		{
			get { return _errorKind; }
		}

		public object Sender
		{
			get { return _origSender; }
		}

		public ValidationEventArgs Arguments
		{
			get { return _valArgs; }
		}
	}

	public delegate void XMLErrorHandler(object sender, ErrorEventArgs vargs);
}
