using System;
using Rail.Reflect;

namespace AspectNet
{
	#region WeaveContext
	public class WeaveContext
	{
		public JoinPoint	pointToTransform	= null;
		public Advice		targetAdvice		= null;

		public RAssemblyDef	pointAssembly		= null;
		public RModuleDef	pointModule			= null;
		public RTypeDef		pointType			= null;
		public RMethodDef	pointMethod			= null;
		public int			binding				= -1;

		public static WeaveContext CreateContext( JoinPoint point, Advice advice, int binding )
		{
			WeaveContext ctx = new WeaveContext();

			ctx.pointToTransform = point;
			ctx.targetAdvice = advice; 

			ctx.pointAssembly	= point.DeclaringAssembly;
			ctx.pointModule		= point.DeclaringModule;
			ctx.pointType		= point.DeclaringType;
			ctx.pointMethod		= point.MethodDef;
			ctx.binding			= binding;

			return ctx;
		}

		public RType AspectType
		{
			get
			{
				RType res = null;
				RType aspType = targetAdvice.Aspect.AspectType;

				if ( aspType.Assembly.Name.FullName == pointAssembly.Name.FullName )
				{
					res = aspType;
				}
				else
				{
					res = pointAssembly.GetType( aspType.NameWithNameSpace );
				}

				return res;
			}
		}

		public RMethod AdviceMethod
		{
			get
			{
				RMethod meth = targetAdvice.MethodDef;
				return AspectType.GetMethod( meth.Name, meth.ReturnType.ReturnType, meth.GetParameters() );
			}
		}
	}

	#endregion

	/// <summary>
	/// Summary description for IJPWeaver.
	/// </summary>
	public interface IJPWeaver
	{
		/// <summary>
		/// Called to perform the weaving
		/// </summary>
		bool TransformJoinPoint( );

		/// <summary>
		/// Setter for the weaving context
		/// </summary>
		WeaveContext Context { set; }

		/// <summary>
		/// Called to release resources
		/// </summary>
		void Reset();
	}
}
