using System.Collections;
using AspectNetAddIn.UICommands;
using Microsoft.Win32;

namespace AspectNetAddIn
{
	using System;
	using Microsoft.Office.Core;
	using Extensibility;
	using System.Runtime.InteropServices;
	using EnvDTE;

	#region Read me for Add-in installation and setup information.
	// When run, the Add-in wizard prepared the registry for the Add-in.
	// At a later time, if the Add-in becomes unavailable for reasons such as:
	//   1) You moved this project to a computer other than which is was originally created on.
	//   2) You chose 'Yes' when presented with a message asking if you wish to remove the Add-in.
	//   3) Registry corruption.
	// you will need to re-register the Add-in by building the MyAddin21Setup project 
	// by right clicking the project in the Solution Explorer, then choosing install.
	#endregion
	
	/// <summary>
	///   The object for implementing an Add-in.
	/// </summary>
	/// <seealso class='IDTExtensibility2' />
	[GuidAttribute("3B860404-0501-4A4D-A227-CCC7F2753B63"), ProgId("AspectNetAddIn.Connect")]
	public class Connect : Object, Extensibility.IDTExtensibility2, IDTCommandTarget
	{
		private Hashtable _commands = null;

		/// <summary>
		///		Implements the constructor for the Add-in object.
		///		Place your initialization code within this method.
		/// </summary>
		public Connect()
		{
		}

		/// <summary>
		///      Implements the OnConnection method of the IDTExtensibility2 interface.
		///      Receives notification that the Add-in is being loaded.
		/// </summary>
		/// <param term='application'>
		///      Root object of the host application.
		/// </param>
		/// <param term='connectMode'>
		///      Describes how the Add-in is being loaded.
		/// </param>
		/// <param term='addInInst'>
		///      Object representing this Add-in.
		/// </param>
		/// <seealso class='IDTExtensibility2' />
		public void OnConnection(object application, Extensibility.ext_ConnectMode connectMode, object addInInst, ref System.Array custom)
		{
			applicationObject = (_DTE)application;
			addInInstance = (AddIn)addInInst;
			
			InitializeCommands();

			switch(connectMode)
			{
				case(Extensibility.ext_ConnectMode.ext_cm_UISetup):
					InitializeUI();
					break;
				case(Extensibility.ext_ConnectMode.ext_cm_Startup):
					InitializeEvents();
					break;
				default:
					break;
			}
		}

		/// <summary>
		///     Implements the OnDisconnection method of the IDTExtensibility2 interface.
		///     Receives notification that the Add-in is being unloaded.
		/// </summary>
		/// <param term='disconnectMode'>
		///      Describes how the Add-in is being unloaded.
		/// </param>
		/// <param term='custom'>
		///      Array of parameters that are host application specific.
		/// </param>
		/// <seealso class='IDTExtensibility2' />
		public void OnDisconnection(Extensibility.ext_DisconnectMode disconnectMode, ref System.Array custom)
		{
			RegistryKey rk = Registry.CurrentUser.OpenSubKey( "Software\\Microsoft\\VisualStudio\\7.1\\AddIns\\AspectNetAddIn.Connect", true );
			rk.SetValue("CommandPreload", 1);
		}

		/// <summary>
		///      Implements the OnAddInsUpdate method of the IDTExtensibility2 interface.
		///      Receives notification that the collection of Add-ins has changed.
		/// </summary>
		/// <param term='custom'>
		///      Array of parameters that are host application specific.
		/// </param>
		/// <seealso class='IDTExtensibility2' />
		public void OnAddInsUpdate(ref System.Array custom)
		{
		}

		/// <summary>
		///      Implements the OnStartupComplete method of the IDTExtensibility2 interface.
		///      Receives notification that the host application has completed loading.
		/// </summary>
		/// <param term='custom'>
		///      Array of parameters that are host application specific.
		/// </param>
		/// <seealso class='IDTExtensibility2' />
		public void OnStartupComplete(ref System.Array custom)
		{
		}

		/// <summary>
		///      Implements the OnBeginShutdown method of the IDTExtensibility2 interface.
		///      Receives notification that the host application is being unloaded.
		/// </summary>
		/// <param term='custom'>
		///      Array of parameters that are host application specific.
		/// </param>
		/// <seealso class='IDTExtensibility2' />
		public void OnBeginShutdown(ref System.Array custom)
		{
			
		}
		
		/// <summary>
		///      Implements the QueryStatus method of the IDTCommandTarget interface.
		///      This is called when the command's availability is updated
		/// </summary>
		/// <param term='commandName'>
		///		The name of the command to determine state for.
		/// </param>
		/// <param term='neededText'>
		///		Text that is needed for the command.
		/// </param>
		/// <param term='status'>
		///		The state of the command in the user interface.
		/// </param>
		/// <param term='commandText'>
		///		Text requested by the neededText parameter.
		/// </param>
		/// <seealso class='Exec' />
		public void QueryStatus(string commandName, EnvDTE.vsCommandStatusTextWanted neededText, ref EnvDTE.vsCommandStatus status, ref object commandText)
		{
			if (neededText == vsCommandStatusTextWanted.vsCommandStatusTextWantedNone &&
				_commands.ContainsKey( commandName ) )
			{
				BaseCommand cmd = (BaseCommand) _commands[commandName];
				cmd.QueryStatus(commandName,
								neededText,
								ref status,
								ref commandText);
			}
		}

		/// <summary>
		///      Implements the Exec method of the IDTCommandTarget interface.
		///      This is called when the command is invoked.
		/// </summary>
		/// <param term='commandName'>
		///		The name of the command to execute.
		/// </param>
		/// <param term='executeOption'>
		///		Describes how the command should be run.
		/// </param>
		/// <param term='varIn'>
		///		Parameters passed from the caller to the command handler.
		/// </param>
		/// <param term='varOut'>
		///		Parameters passed from the command handler to the caller.
		/// </param>
		/// <param term='handled'>
		///		Informs the caller if the command was handled or not.
		/// </param>
		/// <seealso class='Exec' />
		public void Exec(string commandName, EnvDTE.vsCommandExecOption executeOption, ref object varIn, ref object varOut, ref bool handled)
		{
			handled = false;
			if(executeOption == EnvDTE.vsCommandExecOption.vsCommandExecOptionDoDefault)
			{
				if ( _commands.ContainsKey( commandName ) )
				{
					BaseCommand cmd = (BaseCommand) _commands[commandName];
					cmd.Execute( commandName, executeOption, ref varIn, ref varOut, ref handled );
				}
			}
		}

		/// <summary>
		/// Initializes the commands, needed by the addin
		/// </summary>
		public void InitializeCommands()
		{
			if ( _commands == null )
			{
				_commands = new Hashtable();

				BaseCommand cmd = new TestCommand(this.applicationObject, this._eventSink);
				_commands.Add( addInInstance.ProgID + "." + cmd.Name, cmd );
			}
		}

		/// <summary>
		///      Initializes the commands
		/// </summary>
		private void InitializeUI()
		{
			IDictionaryEnumerator en = _commands.GetEnumerator();
				
			while ( en.MoveNext() )
			{
				BaseCommand cmd = en.Value as BaseCommand;
					
				// first remove the command, thus guarantee single instance
				cmd.RemoveCommand(addInInstance);
					
				//add the command then
				cmd.AddCommand(addInInstance);
			}
		}

		/// <summary>
		///      Initializes the events
		/// </summary>
		private void InitializeEvents()
		{
			if ( !_eventSink.Initialized )
			{
				_eventSink.Initialize( applicationObject );
			}		
		}

		private AspectNetEventSink _eventSink = new AspectNetEventSink();
		private _DTE applicationObject;
		private AddIn addInInstance;
		
	}
}