using System;
using EnvDTE;

namespace AspectNetAddIn
{
	/// <summary>
	/// Summary description for Settings.
	/// </summary>
	internal class Settings
	{
		# region private key constants
		private const string _SCHEMA_LOC_KEY	= "aspectnet_schema_location";
		private const string _OUTPUT_LOC_KEY	= "aspectnet_output_location";
		private const string _ASPECTLIB_LOC_KEY = "aspectnet_aspectlib_location";
		private const string _ENABLED_KEY		= "aspectnet_enabled";

		public const string	 WEAVE_CONFIG_FILE	= "weave_configuration.xml";

		#endregion

		#region public data members
		public string SchemaLocation = @"E:\projects\MSc Thesis\src\AspectNet\aspectnet_schema.xsd";
		public string OutputLocation = @"C:\Documents and Settings\Minimalistic\My Documents\Visual Studio Projects\Solution1\output";
		public string XmlMapping;
		public string AspectLibLocation = @"C:\Documents and Settings\Minimalistic\My Documents\Visual Studio Projects\Solution1\AspectLib.dll";
		public bool Enabled = true;
		#endregion

		private _DTE _application = null;
		

		public Settings(_DTE application)
		{
			_application = application;

			LoadSettings();
		}

		bool CheckCreateVariable(string varName)
		{
			bool bVarExists = _application.Solution.Globals.get_VariableExists( varName );

			if ( !bVarExists )
			{
				_application.Solution.Globals[varName] = "";
				_application.Solution.Globals.set_VariablePersists( varName, true );
			}

			return bVarExists;
		}

		/// <summary>
		/// This property returns if Aspect.Net's
		/// configuration is set up properly
		/// </summary>
		public bool IsConfigurationValid
		{
			get
			{
				return	XmlMapping.Length > 0			&&
					AspectLibLocation.Length > 0	&&
					SchemaLocation.Length > 0		&&
					OutputLocation.Length > 0;
			}
		}

		#region (de)serialization
		public void LoadSettings()
		{
			// read the mapping info
			ProjectItem pItem = _application.Solution.FindProjectItem( WEAVE_CONFIG_FILE );
			
			if ( pItem != null )
			{
				XmlMapping = pItem.get_FileNames(1);	
			}

			;

			if ( !_application.Globals.get_VariableExists( _ENABLED_KEY ) )
			{
				_application.Globals[_ENABLED_KEY] = (true).ToString();
				_application.Globals.set_VariablePersists( _ENABLED_KEY, true );
			}
			else
			{
				Enabled = Boolean.Parse( (string)_application.Globals[_ENABLED_KEY] );
			}


			if ( CheckCreateVariable( _SCHEMA_LOC_KEY ) )
			{
				SchemaLocation = (string)_application.Solution.Globals[_SCHEMA_LOC_KEY];
			}
			
			if ( CheckCreateVariable( _OUTPUT_LOC_KEY ) )
			{
				OutputLocation = (string)_application.Solution.Globals[_OUTPUT_LOC_KEY];
			}

			if ( CheckCreateVariable( _ASPECTLIB_LOC_KEY ) )
			{
				AspectLibLocation = (string)_application.Solution.Globals[_ASPECTLIB_LOC_KEY];
			}
		}

		public void SaveSettings()
		{
			_application.Solution.Globals[_ASPECTLIB_LOC_KEY] = AspectLibLocation ;
			_application.Solution.Globals[_OUTPUT_LOC_KEY] = OutputLocation;
			_application.Solution.Globals[_SCHEMA_LOC_KEY] = SchemaLocation;
			_application.Globals[_SCHEMA_LOC_KEY] = Enabled.ToString();
		}

		#endregion
	};
}
