using System;
using EnvDTE;
using Microsoft.Office.Core;

namespace AspectNetAddIn.UICommands
{
	/// <summary>
	/// Summary description for BaseCommand.
	/// </summary>
	public abstract class BaseCommand
	{
		protected EnvDTE._DTE application;

		protected AspectNetEventSink eventSink;
		protected readonly string commandName;
		
		protected string commandText;
		protected string commandToolTip;
		protected int commandStatus = (int)vsCommandStatus.vsCommandStatusSupported + 
										(int)vsCommandStatus.vsCommandStatusEnabled;

		protected readonly string menubarName;
		protected readonly string menubarCaption;

		/// <summary>
		/// Constructor
		/// </summary>
		public BaseCommand(EnvDTE._DTE _application, AspectNetEventSink _eventSink, string _commandName)
		{
			application = _application;
			commandName = _commandName;
			menubarName = "AspectNetMenuBar";
			menubarCaption = "Aspect.Net";
			eventSink = _eventSink;
		}

		public BaseCommand(EnvDTE._DTE _application, string _commandName, string _menubarName, string _menubarCaption)
		{
			application = _application;
			commandName = _commandName;
			menubarName = _menubarName;
			menubarCaption = _menubarCaption;
		}

		/// <summary>
		/// Provides default logic for removing a command from the IDE
		/// </summary>
		public virtual bool RemoveCommand(AddIn addinInstance)
		{
			bool result = true;

			try
			{
				Commands cmds = application.Commands;
				int id = 0;
				string name = addinInstance.ProgID + "." + Name;
				Command cmd = cmds.Item(name, id);
				cmd.Delete();
			}
			catch
			{
				result = false;
			}

			return result;
		}

		/// <summary>
		/// Provides default logic for adding the command to the IDE
		/// </summary>
		public virtual void AddCommand(AddIn addinInstance)
		{
			object [] contextGUIDs = new object[] {};
			
			Command cmd = application.Commands.AddNamedCommand( addinInstance, 
																commandName, 
																commandText,
																commandToolTip, 
																false,
																-1,
																ref contextGUIDs,
																commandStatus );

			cmd.AddControl(MenuBar.CommandBar, 1);
		}

		/// <summary>
		/// Ensure the menubar exists and returns its instance
		/// </summary>
		protected virtual CommandBarPopup MenuBar
		{
			get
			{
				CommandBar mBar = application.CommandBars["MenuBar"];
				CommandBarControl toolsControl = mBar.Controls["Tools"];
			
				CommandBarPopup popup = null;

				try
				{
					popup = (CommandBarPopup)mBar.Controls[menubarName];
				}
				catch
				{
					popup = null;
				}

				if ( popup == null )
				{
					popup = mBar.Controls.Add( MsoControlType.msoControlPopup,
												Type.Missing,
												Type.Missing,
												mBar.Controls["Help"].Index,
												true ) as CommandBarPopup;

					popup.CommandBar.Name = menubarName;
					popup.Caption = menubarCaption;
					popup.Visible = true;
				}
				
				return popup;
			}
		}

		/// <summary>
		/// Returns the name of the command object
		/// </summary>
		public string Name
		{
			get{ return commandName; }
		}

		/// <summary>
		///      This is called when the command's availability is updated
		/// </summary>
		public abstract void QueryStatus( string commandName,
										  EnvDTE.vsCommandStatusTextWanted neededText, 
										  ref EnvDTE.vsCommandStatus status, 
										  ref object commandText );

		/// <summary>
		///      This is called when the command is invoked.
		/// </summary>
		public abstract void Execute( string commandName, 
			 		  				  EnvDTE.vsCommandExecOption executeOption, 
									  ref object varIn, 
									  ref object varOut, 
									  ref bool handled );
	}
}
