/*
	The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); 
	you may not use this file except in compliance with the License. You may obtain a copy of the License at 
	http://www.mozilla.org/MPL/ 
	Software distributed under the License is distributed on an "AS IS" basis, 
	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the specific language 
	governing rights and limitations under the License. 

	The Original Code is RAIL(Runtime Assembly Instrumentation Library) Alpha Version.
	The Initial Developer of the Original Code is University of Coimbra,
	Computer Science Department, Dependable Systems Group. Copyright (C) University of Coimbra. 
	All Rights Reserved.  
*/
using System;
using System.Reflection;
using System.Collections;
using Rail.Transformation;

namespace Rail.Reflect
{
	/// <summary>
	/// 
	/// </summary>
	public abstract class RProperty : RMember	
	{
		#region fields
		/// <summary>
		/// 
		/// </summary>
		protected PropertyAttributes attributes;
		/// <summary>
		/// 
		/// </summary>
		protected RMethod getMethod;
		/// <summary>
		/// 
		/// </summary>
		protected RMethod setMethod;
		/// <summary>
		/// 
		/// </summary>
		protected RType propertyType;
		/// <summary>
		/// 
		/// </summary>
		protected ArrayList indexParameters;
		#endregion
		
		/// <summary>
		/// 
		/// </summary>
		public virtual PropertyAttributes Attributes 
		{
			get 
			{
				return this.attributes;
			}
			set 
			{
				throw new NotSupportedException();
			}
		}

		/// <summary>
		/// 
		/// </summary>
		public virtual bool CanRead 
		{
			get 
			{
				if (this.getMethod!=null)
					return true;
				return false;
			}      
		}

		/// <summary>
		/// 
		/// </summary>
		public virtual bool CanWrite 
		{
			get 
			{
				if (this.setMethod!=null)
					return true;
				return false;
			}      
		}

		/// <summary>
		/// 
		/// </summary>
		public virtual RType PropertyType 
		{
			get 
			{
				return this.propertyType;
			}
			set 
			{
				throw new NotSupportedException();
			}
		}

		#region Accessors
		/// <summary>
		/// Gets or sets the get accessor method. 
		/// Returns a reference to the internal method object. Changes done 
		/// on that reference will reflect in the internal method. If the method 
		/// name is changed from the default set_/<property/>, an exception 
		/// will be raised when the assembly is generated.
		/// </summary>
		public virtual RMethod GetMethod 
		{
			get 
			{
				return this.getMethod;
			}
			set 
			{
				throw new NotSupportedException();
			}
		}

		/// <summary>
		/// Gets or sets the set accessor method. 
		/// 
		/// Returns a reference to 
		/// the internal method object. Changes done on that reference will
		/// reflect in the internal method. If the method name is changed 
		/// from the default set_/<property/>, an exception will be raised
		/// when the assembly is generated.
		/// </summary>
		public virtual RMethod SetMethod 
		{
			get 
			{
				return this.setMethod;
			}
			set 
			{
				throw new NotSupportedException();
			}
		}
		#endregion

		#region Parameters
		/// <summary>
		/// Only the default property or indexer (refered as Item) has more than
		/// one attribute.
		/// </summary>
		/// <returns></returns>
		public virtual RParameter[] GetIndexParameters() 
		{      
			if (this.indexParameters!=null)
			{
				IEnumerator ienum = this.indexParameters.GetEnumerator();
				RParameter [] rps = new RParameter[this.indexParameters.Count];
				int ipIndex = 0;
				while (ienum.MoveNext())
				{
					rps[ipIndex] = (RParameter)ienum.Current;
					ipIndex++;
				}
				return rps;
			}
			return new RParameter[0];
		}
		#endregion
	}

	/// <summary>
	/// Done
	/// </summary>
	public class RPropertyRef : RProperty 
	{
		#region fields
		/// <summary>
		/// 
		/// </summary>
		private RAssemblyDef rad;
		#endregion
		/// <summary>
		/// 
		/// </summary>
		/// <param name="pi"></param>
		/// <param name="rad"></param>
		internal RPropertyRef(PropertyInfo pi,RAssemblyDef rad) 
		{			
			this.rad = rad;
			this.attributes = pi.Attributes;
			this.declaringType = rad.GetRTypeRef(pi.DeclaringType);
			MethodInfo mi = pi.GetGetMethod();
			if (mi!=null)
				this.getMethod = new RMethodRef(mi,rad);
			mi = null;
			mi = pi.GetSetMethod();
			if (mi!=null)
				this.setMethod = new RMethodRef(mi,rad);
			ParameterInfo [] pis = pi.GetIndexParameters();
			this.indexParameters = new ArrayList();
			if (pis!=null)
				for (int i = 0; i < pis.Length; i++)
					this.indexParameters.Add(new RParameter(pis[i],rad));
			this.memberType = pi.MemberType;
			this.name = pi.Name;
			this.propertyType = rad.GetRTypeRef(pi.PropertyType);
		}
	}

	/// <summary>
	/// Done
	/// </summary>
	public class RPropertyDef : RProperty, IVisitable, ICustomAttributeOwnerDef 
	{
		/// <summary>
		/// 
		/// </summary>
		internal RPropertyDef() 
		{			
			this.indexParameters = new ArrayList();
		}

		/// <summary>
		/// 
		/// </summary>
		public override PropertyAttributes Attributes 
		{     
			set 
			{
				this.attributes = value;
			}
		}

		/// <summary>
		/// 
		/// </summary>
		public override RType PropertyType 
		{
			set 
			{
				this.propertyType = value;
			}
		}

		/// <summary>
		/// 
		/// </summary>
		public override String Name 
		{
			set 
			{
				this.name = value;
			}
		}

		#region Accessors
		/// <summary>
		/// Gets or sets the get accessor method. 
		/// Returns a reference to the internal method object. Changes done 
		/// on that reference will reflect in the internal method. If the method 
		/// name is changed from the default set_/<property/>, an exception 
		/// will be raised when the assembly is generated.
		/// </summary>
		public override RMethod GetMethod 
		{      
			set 
			{
				this.getMethod = value;
			}
		}

		/// <summary>
		/// Gets or sets the set accessor method. 
		/// 
		/// Returns a reference to 
		/// the internal method object. Changes done on that reference will
		/// reflect in the internal method. If the method name is changed 
		/// from the default set_/<property/>, an exception will be raised
		/// when the assembly is generated.
		/// </summary>
		public override RMethod SetMethod 
		{      
			set 
			{
				this.setMethod = value;
			}
		}
		#endregion

		#region Parameters    
		/// <summary>
		/// 
		/// </summary>
		/// <param name="index"></param>
		/// <returns></returns>
		public RParameter RemoveIndexParameter(int index) 
		{
			if (this.indexParameters==null) 
				return null;
			if (this.indexParameters.Count<=index)
				throw new NullReferenceException("Out of bounds");
			IEnumerator ienum = this.indexParameters.GetEnumerator();
			RParameter retParam;
			int ipIndex = 0;
			while (ipIndex<=index)
				ienum.MoveNext();
			retParam = (RParameter)ienum.Current;
			this.indexParameters.RemoveAt(index);
			return retParam;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="parameter"></param>
		/// <param name="index"></param>
		public void InsertIndexParameter(RParameter parameter, int index) 
		{
			if (this.indexParameters==null)
				this.indexParameters = new ArrayList();
			if (index==0)
				throw new InvalidOperationException();
			parameter.Sequence = index;
			this.indexParameters.Insert(index,parameter);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="parameter"></param>
		public void SetIndexParameters(RParameter [] parameter) 
		{
			if (parameter!=null)
			{
				this.indexParameters = new ArrayList();
				for (int i = 0; i<parameter.Length;i++)
					this.indexParameters.Add(parameter[i]);
			}
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="parameter"></param>
		public void SetIndexParameters(RType [] parameter) 
		{
			if (parameter!=null)
			{
				this.indexParameters = new ArrayList();
				for (int i = 0; i<parameter.Length;i++)
					this.indexParameters.Add(new RParameter(i+1,parameter[i]));
			}
		}

		#endregion

        	
		#region Custom Attributes
		/// <summary>
		/// 
		/// </summary>
		/// <param name="attribute"></param>
		public void AddCustomAttributtes(RCustomAttrib attribute) 
		{
			InternalAddCustomAttributtes(attribute);
		}

		/// <summary>
		/// TODO: How should the attribute be specified?
		/// </summary>
		/// <param name="attribute"></param>
		public void RemoveCustomAttribute(RCustomAttrib attribute) 
		{
			InternalRemoveCustomAttribute(attribute);
		}

		#endregion
		/// <summary>
		/// Visits this property.
		/// </summary>
		/// <remarks>
		/// A property has no subelements    
		/// </remarks>
		/// <param name="v">The visitor to be used.</param>
		public void Accept(Visitor v) 
		{
			v.GetWalker().WalkProperty(this);
		}
	}
}
