/*
	The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); 
	you may not use this file except in compliance with the License. You may obtain a copy of the License at 
	http://www.mozilla.org/MPL/ 
	Software distributed under the License is distributed on an "AS IS" basis, 
	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the specific language 
	governing rights and limitations under the License. 

	The Original Code is RAIL(Runtime Assembly Instrumentation Library) Alpha Version.
	The Initial Developer of the Original Code is University of Coimbra,
	Computer Science Department, Dependable Systems Group. Copyright (C) University of Coimbra. 
	All Rights Reserved.  
*/
using System;

namespace Rail.Exceptions
{
	/// <summary>
	/// The caracterizes the type of the handler and the handler code to Add
	/// </summary>
	public enum HandlerCode
	{
		CatchAndLog = 0x01
	}
	
	/// <summary>
	/// The Custom attribute to use in the EH functionality
	/// </summary>
	[AttributeUsage(AttributeTargets.Method,
		 AllowMultiple=true,
		 Inherited=false)]
	public class ExceptionAttribute : System.Attribute
	{
		private Type exceptionType;

		public ExceptionAttribute(Type exceptionType)
		{
			this.exceptionType = exceptionType;
		}
		
		public Type ExceptionType
		{
			get
			{
				return this.exceptionType;
			}
		}
	}

	/// <summary>
	/// The marking method to the begin and end of the EH clauses
	/// </summary>
	public class Marker
	{
		public static void Start()
		{
		}
		public static void End()
		{
		}
	}
	/// <summary>
	/// 
	/// </summary>
	public class InternalErrorException : ApplicationException
	{
		/// <summary>
		/// 
		/// </summary>
		/// <param name="error"></param>
		public InternalErrorException(String error) 
      : base("[Internal error] " + error)
		{     
		}
	}
}
