/*
	The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); 
	you may not use this file except in compliance with the License. You may obtain a copy of the License at 
	http://www.mozilla.org/MPL/ 
	Software distributed under the License is distributed on an "AS IS" basis, 
	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the specific language 
	governing rights and limitations under the License. 

	The Original Code is RAIL(Runtime Assembly Instrumentation Library) Alpha Version.
	The Initial Developer of the Original Code is University of Coimbra,
	Computer Science Department, Dependable Systems Group. Copyright (C) University of Coimbra. 
	All Rights Reserved.  
*/
using System;
using Rail;
using Rail.Reflect;
using System.Collections;


namespace Rail.Transformation
{

	/// <summary>
	/// Applies code transformations to the methods (<code>RMethodDef</code>) 
	/// and constructors (<code>RConstructorDef</code>) visited.
	/// <remarks>
	/// This instances of this class save a list of transformations that can be 
	/// applied to MSIL code, and are only efectively applied when the instance 
	/// of the class is passed to the method or methods through the <code>Accept</code>
	/// method in the destination object
	/// </remarks>
	/// </summary>
	sealed public class CodeTransformer : EmptyVisitor 
	{
		/// <summary>
		/// An <code>System.Collections.Arraylist</code> object used to store the
		/// list of transformations to apply to the methods and constructors
		/// <seealso cref="Rail.Transformer"/>
		/// <seealso cref="Rail.Transformers"/>
		/// </summary>
		private ArrayList Conversions;

		/// <summary>
		/// A <code>CodeTransformer</code> defines transformations to be applied
		/// to code.
		/// </summary>
		public CodeTransformer() 
		{	
			this.Conversions = new ArrayList();
		}
    

		#region Method's instrumentation
		/// <summary>
		/// Adds a new visitor (transformer) to the collection owned by this object
		/// </summary>
		/// <param name="codeVisitor">The visitor to be inserted</param>
		public void add(Visitor codeVisitor) 
		{
			this.Conversions.Add(codeVisitor);
		}    
		#endregion

		#region Visitor Implementation
		/// <summary>
		/// 
		/// </summary>
		/// <param name="method"></param>
		public override void VisitConstructor(RConstructorDef method) 
		{
			foreach(Visitor conversion in this.Conversions)
				conversion.VisitConstructor(method);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="method"></param>
		public override void VisitMethod(RMethodDef method) 
		{    
			foreach(Visitor conversion in this.Conversions)
				conversion.VisitMethod(method);
		}


		#endregion
	}

	
}
