/*
	The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); 
	you may not use this file except in compliance with the License. You may obtain a copy of the License at 
	http://www.mozilla.org/MPL/ 
	Software distributed under the License is distributed on an "AS IS" basis, 
	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the specific language 
	governing rights and limitations under the License. 

	The Original Code is RAIL(Runtime Assembly Instrumentation Library) Alpha Version.
	The Initial Developer of the Original Code is University of Coimbra,
	Computer Science Department, Dependable Systems Group. Copyright (C) University of Coimbra. 
	All Rights Reserved.  
*/
using System;
using Rail.Reflect;
using System.Reflection;
using System.Reflection.Emit;
using System.Collections;

namespace Rail.Transformation
{
	/// <summary>
	/// Useful to change method access from a method to other inside the method body 
	/// <remarks> Both methods must have the same number, order and type of arguments.
	/// Both methods have to be static or both have to be defined in the same type</remarks>
	/// </summary>
	public class MethodReplacer : EmptyVisitor
	{
		/// <summary>
		/// 
		/// </summary>
		private readonly RMethod oldMethod, newMethod;

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="oldMethod">The method to be replaced</param>
		/// <param name="newMethod">The method to use instead</param>
		public MethodReplacer(RMethod oldMethod, RMethod newMethod)
		{
			this.oldMethod = oldMethod;
			this.newMethod = newMethod;
		}
        
		/// <summary>
		/// Method used to instrument the code of <code>RMethodDef</code> and 
		/// <code>RConstructorDef</code> 
		/// </summary>
		/// <param name="method">The method which method body is to be instrumented</param>
		private void VisitMethodBody(RMethodBase method)
		{
			
			#region Changes the body of the method
			
			MSIL.MethodBody mb = null;
			if (method is RConstructor)
				mb = ((RConstructorDef)method).MethodBody;
			else if (method is RMethod)
				mb = ((RMethodDef)method).MethodBody;

			MSIL.Code code = null;
			if (mb!=null)
				code = mb.GetCode();
			if (code!=null)
				foreach (MSIL.Instruction i in code)
				{
					if (i is MSIL.ILMethod)
					{
						MSIL.ILMethod d = (MSIL.ILMethod)i;
						if (d.Method.Equals(this.oldMethod))
						{
							((MSIL.ILMethod)i).Method = this.newMethod;							
							if (((MSIL.ILMethod)i).Method==null)
								throw new NullReferenceException();
						}
					}
					else if (i is MSIL.ILToken)
					{
						MSIL.ILToken e = (MSIL.ILToken)i;
						if (e.Token is RMethod)
						{
							RMethod rm = e.Token as RMethod;
							if (rm.Equals(this.oldMethod))
							{
								((MSIL.ILToken)i).Token = this.newMethod;
								if (((MSIL.ILToken)i).Token==null)
									throw new NullReferenceException();
							}
						}
					}
				}
			
			#endregion
			
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="method"></param>
		public override void VisitMethod(RMethodDef method) 
		{
			//check the method body
			VisitMethodBody((RMethodBase)method);
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="method"></param>
		public override void VisitConstructor(RConstructorDef method) 
		{
			//Check the method body
			VisitMethodBody((RMethodBase)method);
		}


	}
}
