using System;
using System.Collections;
using System.Xml.Schema;
using AspectNet;
using AspectNet.AOP;
using Rail.Reflect;

namespace WeaveTester
{
	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	class WeaveTester
	{
		private static string assembly_name = "testassembly.dll";
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			string folder = @"E:\projects\MSc Thesis\src";

			// =================================================================================================================
			// identify joinpoints
			PointCutManager pcmngr = new PointCutManager( @"E:\projects\MSc Thesis\src\AspectNet\aspectnet_schema.xsd" );
			AspectWeaver wv = new AspectWeaver();
			wv.OutputFolder = folder + @"\output";
			
			// in a loop for each built assembly, we should call both - weaver and the PC manager
			// to:
			//		1. adjust the aspect classes
			//		2. collect assemblies paths, for further JP identification
			wv.LoadAspects( folder + @"\WeaveTester\bin\Debug\TestAssembly.dll" );
			pcmngr.ProcessAssembly( folder + @"\Weave\bin\Debug\TestAssembly.dll" );

			
			// now make sure that we handle the case in which we have aspects and base coding
			// (affected by the aspects) in the same assembly
			pcmngr.UpdateAssemblyPaths( wv.AdjustedAssemblies );

			// now do the real joinpoints identification
			pcmngr.XmlMappingPath = folder + @"\TestApplication\TestMapping.xml";
			pcmngr.BuildJoinPoints();

			// now do the swing :)
			wv.ProcessJoinPoints( pcmngr.PointCuts );



			// =================================================================================================================
			// display results

			if ( pcmngr.PointCuts.Count > 0 )
			{
				ArrayList res = (pcmngr.PointCuts[0] as ArrayList);

				res.Sort(new Comparator());

				foreach ( JoinPoint rm in res )
				{
					System.Console.WriteLine( rm.DeclaringType.Name 
						+ "." 
						+ rm.MethodDef.Name 
						+ ":" 
						+ rm.MethodDef.ParametersCount.ToString());

					foreach (RParameter rp in rm.MethodDef.GetParameters())
					{
						System.Console.WriteLine("\t" + rp.Name + "[" + rp.ParameterType.NameWithNameSpace + "]" );
					}
				}
			}
		}

		private class Comparator : IComparer
		{
			#region IComparer Members

			public int Compare(object jp1, object jp2)
			{
				RMethodDef x = (jp1 as JoinPoint).MethodDef;
				RMethodDef y = (jp2 as JoinPoint).MethodDef;

				string str1 = x.DeclaringType.Name + "." + 
					x.Name + ":" + 
					x.ParametersCount.ToString();
				string str2 = y.DeclaringType.Name + "." + 
					y.Name + ":" + 
					y.ParametersCount.ToString();
				
				return str1.CompareTo(str2);
			}

			#endregion

		}

		protected void alabala()
		{
			
		}

		public static void ValidationHandler(object sender, ValidationEventArgs args)
		{
			System.Diagnostics.Trace.WriteLine("***Validation error");
			System.Diagnostics.Trace.WriteLine(string.Format("\tSeverity:{0}", args.Severity));
			System.Diagnostics.Trace.WriteLine(string.Format("\tMessage:{0}", args.Message));
		}

		static void ProcessAssembly()
		{
			AspectWeaver wv = new AspectWeaver();
			
			wv.LoadAspects( assembly_name );
		}
	}
}
